/*
 * G-code meta compiler
 *
 * Copyright (C) 2013  B. Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <getopt.h>
#include <limits.h>
#include <ctype.h>

#include "utils.h"
#include "value.h"
#include "variable.h"
#include "builtin.h"

/*#define FREEONEXIT	1*/	/* For valgrind testing */

#define WARNSHIFTS	1000	/* Trying to shift this many times may be an error */

static value_t *execute(node_t *head, node_t **brkref);
static value_t *interpret(node_t *n, node_t **brkref);
static void interpret_expr(node_t *n);

FILE *ofp;
double *global_offs[MAXAXES];
double *global_pos[MAXAXES];
plane_et global_plane = PLANE_XY;
int naxes = 6;			/* Default enable 6 axes */
int cl_decimals = 8;		/* Number of significant decimals to print in output */
int cl_relative;		/* Use relative output code */
int cl_inch;			/* Set if output is to be in inch */
int runtimeerrors;		/* Nr of calls to script's error() function */
int runtimewarnings;		/* Nr of calls to script's warning() function or internal rtwarning() */

static node_t **functions;
static int nfunctions;
static int nafunctions;

static int syntaxerrors;

int yyparse(void);

typedef value_t *(*opfunc_t)(const node_t *ref, value_t *lv, const value_t *rv);

static int errorprint(const node_t *n, const char *fmt, const char *pfx, va_list va)
{
	int l = linenr;
	int c = charnr;
	if(n) {
		l = n->linenr;
		c = n->charnr;
	}
	fprintf(stderr, "%s:%d:%d: %s: ", filename, l, c, pfx);
	vfprintf(stderr, fmt, va);
	fprintf(stderr, "\n");
	return 0;
}

int yyerror(const char *fmt, ...)
{
	syntaxerrors++;
	/* The parser always emits "syntax error" before going into error-recovery */
	if(!strcmp(fmt, "syntax error"))
		return 0;
	va_list va;
	va_start(va, fmt);
	errorprint(NULL, fmt, "error", va);
	va_end(va);
	return 0;
}

int yyfatal(const char *fmt, ...)
{
	va_list va;
	va_start(va, fmt);
	errorprint(NULL, fmt, "fatal", va);
	va_end(va);
	exit(1);
}

int rtinternal(const node_t *n, const char *fmt, ...)
{
	va_list va;
	va_start(va, fmt);
	errorprint(n, fmt, "internal error", va);
	va_end(va);
	exit(1);
}

int rterror(const node_t *n, const char *fmt, ...)
{
	va_list va;
	va_start(va, fmt);
	errorprint(n, fmt, "error", va);
	va_end(va);
	exit(1);
}

int rtwarning(const node_t *n, const char *fmt, ...)
{
	va_list va;
	va_start(va, fmt);
	errorprint(n, fmt, "warning", va);
	va_end(va);
	runtimewarnings++;
	return 0;
}

static const char *nodetypename(node_et n)
{
	switch(n) {
	case NT_INVALID:	return "NT_INVALID";
	case NT_EXPR:		return "NT_EXPR";
	case NT_EXPRLIST:	return "NT_EXPRLIST";
	case NT_IF:		return "NT_IF";
	case NT_FOR:		return "NT_FOR";
	case NT_FOREACH:	return "NT_FOREACH";
	case NT_WHILE:		return "NT_WHILE";
	case NT_DOWHILE:	return "NT_DOWHILE";
	case NT_FUNCTION:	return "NT_FUNCTION";
	case NT_ARGLIST:	return "NT_ARGLIST";
	case NT_CALL:		return "NT_CALL";
	case NT_BREAK:		return "NT_BREAK";
	case NT_CONTINUE:	return "NT_CONTINUE";
	case NT_RETURN:		return "NT_RETURN";
	case NT_LOCAL:		return "NT_LOCAL";
	}
	return "<unknown node type>";
}

static const char *expropname(op_et n)
{
	switch(n) {
	case OP_NULL:		return "OP_NULL";
	case OP_ADD:		return "OP_ADD";
	case OP_SUB:		return "OP_SUB";
	case OP_MUL:		return "OP_MUL";
	case OP_DIV:		return "OP_DIV";
	case OP_MOD:		return "OP_MOD";
	case OP_INDEX:		return "OP_INDEX";
	case OP_LOR:		return "OP_LOR";
	case OP_LAND:		return "OP_LAND";
	case OP_EQ:		return "OP_EQ";
	case OP_NE:		return "OP_NE";
	case OP_GT:		return "OP_GT";
	case OP_LT:		return "OP_LT";
	case OP_GE:		return "OP_GE";
	case OP_LE:		return "OP_LE";
	case OP_SHL:		return "OP_SHL";
	case OP_SHR:		return "OP_SHR";
	case OP_NOT:		return "OP_NOT";
	case OP_INT:		return "OP_INT";
	case OP_FLOAT:		return "OP_FLOAT";
	case OP_STRING:		return "OP_STRING";
	case OP_VECTOR:		return "OP_VECTOR";
	case OP_VECTORLIST:	return "OP_VECTORLIST";
	case OP_ASSIGN:		return "OP_ASSIGN";
	case OP_ADDASSIGN:	return "OP_ADDASSIGN";
	case OP_SUBASSIGN:	return "OP_SUBASSIGN";
	case OP_MULASSIGN:	return "OP_MULASSIGN";
	case OP_DIVASSIGN:	return "OP_DIVASSIGN";
	case OP_MODASSIGN:	return "OP_MODASSIGN";
	case OP_SHLASSIGN:	return "OP_SHLASSIGN";
	case OP_SHRASSIGN:	return "OP_SHRASSIGN";
	case OP_INC:		return "OP_INC";
	case OP_DEC:		return "OP_DEC";
	case OP_DEREF:		return "OP_DEREF";
	}
	return "<unknown expression operation>";
}

static const char *unitname(int n)
{
	switch(n) {
	case UNIT_NONE:	return "UNIT_NONE";
	case UNIT_MM:	return "UNIT_MM";
	case UNIT_IN:	return "UNIT_IN";
	case UNIT_DEG:	return "UNIT_DEG";
	case UNIT_RAD:	return "UNIT_RAD";
	}
	return "<unknown unit>";
}

static void unit_check(const node_t *n, const value_t *v)
{
	switch(v->unit) {
	case UNIT_NONE:
	case UNIT_MM:
	case UNIT_IN:
	case UNIT_DEG:
	case UNIT_RAD:
		break;
	default:
		rtinternal(n, "Invalid unit value %d", v->unit);
	}
}

static value_t *op_add(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	double dval;
	int unitmismatch = 0;
	assert(lv != NULL);
	assert(rv != NULL);

	if(rv->type == VAL_UNDEF) {
		/* x + "undef" = x + 0 = x */
		return lv;
	}
	if(lv->type == VAL_UNDEF) {
		if(!isscalar(rv)) {
			rterror(ref, "Operation '<undef> + non-scalar' not supported");
			return lv;
		}
		/* "undef" + x = 0 + x */
		*lv = *rv;
		return lv;
	}

	/* Type check
	 * scalar + scalar
	 * vector + vector
	 * vectorlist + vector
	 * vectorlist + vectorlist	(append)
	 */
	if(!((isscalar(lv) && isscalar(rv)) || (isvector(lv) && isvector(rv)) || (isvectorlist(lv) && (isvector(rv) || isvectorlist(rv))))) {
		rterror(ref, "Scalars and vectors can only be added to scalars vectors and vector lists (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(isangle(lv) && isdistance(rv)) {
		rtwarning(ref, "Operation '+': LHS has angular units and RHS has distance units, RHS units ignored");
		unitmismatch = 1;
	} else if(isdistance(lv) && isangle(rv)) {
		rtwarning(ref, "Operation '+': LHS has distance units and RHS has angular units, RHS units ignored");
		unitmismatch = 1;
	}

	if(lv->type == VAL_INT && rv->type == VAL_FLOAT) {
		lv->d = lv->i;
		lv->type = VAL_FLOAT;
	}

	switch(lv->type) {
	case VAL_INT:
		assert(isscalar(rv));
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->i += rv->i;
		} else if(lv->unit == UNIT_NONE) {
			lv->i += rv->i;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = (double)lv->i + (double)rv->i * 25.4;
			else
				lv->d = (double)lv->i + (double)rv->i / 25.4;
			lv->type = VAL_FLOAT;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = (double)lv->i + (double)rv->i * M_PI / 180.0;
			else
				lv->d = (double)lv->i + (double)rv->i / M_PI * 180.0;
			lv->type = VAL_FLOAT;
		} else
			rtinternal(ref, "Unit mismatch in op_add %d, %d", lv->unit, rv->unit);
		break;
	case VAL_FLOAT:
		assert(isscalar(rv));
		if(rv->type == VAL_INT)
			dval = rv->i;
		else
			dval = rv->d;
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->d += dval;
		} else if(lv->unit == UNIT_NONE) {
			lv->d += dval;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d += dval * 25.4;
			else
				lv->d += dval / 25.4;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d += dval * M_PI / 180.0;
			else
				lv->d += dval / M_PI * 180.0;
		} else
			rtinternal(ref, "Unit mismatch in op_add %d, %d", lv->unit, rv->unit);
		break;
	case VAL_VECTOR:
		while(lv->v.n < rv->v.n)
			value_vector_add_undef(lv);
		for(i = 0; i < lv->v.n && i < rv->v.n; i++)
			op_add(ref, lv->v.vals[i], rv->v.vals[i]);
		break;
	case VAL_VECTORLIST:
		if(rv->type == VAL_VECTOR) {
			for(i = 0; i < lv->vl.n; i++)
				op_add(ref, lv->vl.vecs[i], rv);
		} else {
			assert(rv->type == VAL_VECTORLIST);
			/* Append vectorlist, transfer from rv to lv */
			for(i = 0; i < rv->vl.n; i++)
				value_vectorlist_add(lv, value_dup(rv->vl.vecs[i]));
		}
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_add %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_sub(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	double dval;
	int unitmismatch = 0;
	assert(lv != NULL);
	assert(rv != NULL);

	if(rv->type == VAL_UNDEF) {
		/* x - "undef" = x - 0 = x */
		return lv;
	}
	if(lv->type == VAL_UNDEF) {
		/* "undef" - x = 0 - x */
		lv->type = VAL_INT;
		lv->i = 0;
		lv->unit = UNIT_NONE;
	}

	/* Type check
	 * scalar - scalar
	 * vector - vector
	 * vectorlist - vector
	 */
	if(!((isscalar(lv) && isscalar(rv)) || ((isvector(lv) || isvectorlist(lv)) && isvector(rv)))) {
		rterror(ref, "Only scalars/vectors can be subtracted from scalars/vectors (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}
	unit_check(ref, lv);
	unit_check(ref, rv);

	if(isangle(lv) && isdistance(rv)) {
		rtwarning(ref, "Operation '-': LHS has angular units and RHS has distance units, RHS units ignored");
		unitmismatch = 1;
	} else if(isdistance(lv) && isangle(rv)) {
		rtwarning(ref, "Operation '-': LHS has distance units and RHS has angular units, RHS units ignored");
		unitmismatch = 1;
	}

	if(lv->type == VAL_INT && rv->type == VAL_FLOAT) {
		lv->d = lv->i;
		lv->type = VAL_FLOAT;
	}

	switch(lv->type) {
	case VAL_INT:
		assert(isscalar(rv));
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->i -= rv->i;
		} else if(lv->unit == UNIT_NONE) {
			lv->i -= rv->i;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = (double)lv->i - (double)rv->i * 25.4;
			else
				lv->d = (double)lv->i - (double)rv->i / 25.4;
			lv->type = VAL_FLOAT;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = (double)lv->i - (double)rv->i * M_PI / 180.0;
			else
				lv->d = (double)lv->i - (double)rv->i / M_PI * 180.0;
			lv->type = VAL_FLOAT;
		} else
			rtinternal(ref, "Unit mismatch in op_sub %d, %d", lv->unit, rv->unit);
		break;
	case VAL_FLOAT:
		assert(isscalar(rv));
		if(rv->type == VAL_INT)
			dval = rv->i;
		else
			dval = rv->d;
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->d -= dval;
		} else if(lv->unit == UNIT_NONE) {
			lv->d -= dval;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d -= dval * 25.4;
			else
				lv->d -= dval / 25.4;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d -= dval * M_PI / 180.0;
			else
				lv->d -= dval / M_PI * 180.0;
		} else
			rtinternal(ref, "Unit mismatch in op_sub %d, %d", lv->unit, rv->unit);
		break;
	case VAL_VECTOR:
		while(lv->v.n < rv->v.n)
			value_vector_add_undef(lv);
		for(i = 0; i < lv->v.n && i < rv->v.n; i++)
			op_sub(ref, lv->v.vals[i], rv->v.vals[i]);
		break;
	case VAL_VECTORLIST:
		for(i = 0; i < lv->vl.n; i++)
			op_sub(ref, lv->vl.vecs[i], rv);
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_sub %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_mul(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	double dval;
	int unitmismatch = 0;
	assert(lv != NULL);
	assert(rv != NULL);

	/* Multiply by "undef" = "undef" */
	if(lv->type == VAL_UNDEF)
		return lv;
	if(rv->type == VAL_UNDEF) {
		if(!isscalar(lv)) {
			rterror(ref, "Operation 'non-scalar * <undef>' not supported");
			return lv;
		}
		lv->type = VAL_UNDEF;
		return lv;
	}

	/* Type check
	 * scalar * scalar
	 * vector * scalar
	 * vector * vector	(dot product)
	 * vectorlist * scalar	(each vector multiplied by scalar)
	 */
	if(!(	(isscalar(lv) && isscalar(rv)) ||
		(isvector(lv) && (isscalar(rv) || isvector(rv))) ||
		(isvectorlist(lv) && isscalar(rv)))) {
		rterror(ref, "Scalars and vectors can only be multiplied by scalars and vectors (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(isangle(lv) && isdistance(rv)) {
		rtwarning(ref, "Operation '*': LHS has angular units and RHS has distance units, RHS units ignored");
		unitmismatch = 1;
	} else if(isdistance(lv) && isangle(rv)) {
		rtwarning(ref, "Operation '*': LHS has distance units and RHS has angular units, RHS units ignored");
		unitmismatch = 1;
	}

	if(lv->type == VAL_INT && rv->type == VAL_FLOAT) {
		lv->d = lv->i;
		lv->type = VAL_FLOAT;
	}

	switch(lv->type) {
	case VAL_INT:
		assert(isscalar(rv));
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->i *= rv->i;
		} else if(lv->unit == UNIT_NONE) {
			lv->i *= rv->i;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = (double)lv->i * (double)rv->i * 25.4;
			else
				lv->d = (double)lv->i * (double)rv->i / 25.4;
			lv->type = VAL_FLOAT;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = (double)lv->i * (double)rv->i * M_PI / 180.0;
			else
				lv->d = (double)lv->i * (double)rv->i / M_PI * 180.0;
			lv->type = VAL_FLOAT;
		} else
			rtinternal(ref, "Unit mismatch in op_mul %d, %d", lv->unit, rv->unit);
		break;
	case VAL_FLOAT:
		assert(isscalar(rv));
		if(rv->type == VAL_INT)
			dval = rv->i;
		else
			dval = rv->d;
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->d *= dval;
		} else if(lv->unit == UNIT_NONE) {
			lv->d *= dval;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d *= dval * 25.4;
			else
				lv->d *= dval / 25.4;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d *= dval * M_PI / 180.0;
			else
				lv->d *= dval / M_PI * 180.0;
		} else
			rtinternal(ref, "Unit mismatch in op_mul %d, %d", lv->unit, rv->unit);
		break;
	case VAL_VECTOR:
		if(isscalar(rv)) {
			for(i = 0; i < lv->v.n; i++) {
				op_mul(ref, lv->v.vals[i], rv);
			}
		} else {
			/* Vector * vector == scalar (dot product) */
			double sum = 0.0;
			for(i = 0; i < lv->v.n && i < rv->v.n; i++) {
				sum += value_to_mm(lv->v.vals[i]) * value_to_mm(rv->v.vals[i]);
			}
			value_delete_vector(lv);
			lv->type = VAL_FLOAT;
			lv->unit = UNIT_NONE;
			lv->d = sum;
		}
		break;
	case VAL_VECTORLIST:
		assert(isscalar(rv));
		for(i = 0; i < lv->vl.n; i++)
			op_mul(ref, lv->vl.vecs[i], rv);
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_mul %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_div(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	double dval;
	int unitmismatch = 0;
	assert(lv != NULL);
	assert(rv != NULL);

	/* Divide by "undef" = "undef" */
	if(lv->type == VAL_UNDEF)
		return lv;
	if(rv->type == VAL_UNDEF) {
		if(!isscalar(lv)) {
			rterror(ref, "Operation 'non-scalar / <undef>' not supported");
			return lv;
		}
		lv->type = VAL_UNDEF;
		return lv;
	}

	/* Type check
	 * scalar / scalar
	 * vector / scalar
	 * vectorlist / scalar
	 */
	if(!isscalar(rv) || !(isscalar(lv) || isvector(lv) || isvectorlist(lv))) {
		rterror(ref, "Scalars, vectors and vectorlists can only be divided by scalars (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(isangle(lv) && isdistance(rv)) {
		rtwarning(ref, "Operation '/': LHS has angular units and RHS has distance units, RHS units ignored");
		unitmismatch = 1;
	} else if(isdistance(lv) && isangle(rv)) {
		rtwarning(ref, "Operation '/': LHS has distance units and RHS has angular units, RHS units ignored");
		unitmismatch = 1;
	}

	if(lv->type == VAL_INT && rv->type == VAL_FLOAT) {
		lv->d = lv->i;
		lv->type = VAL_FLOAT;
	}

	switch(lv->type) {
	case VAL_INT:
		assert(isscalar(rv));
		if(!rv->i)
			rterror(ref, "Divide by zero");
		if(lv->unit == rv->unit || unitmismatch) {
			lv->i /= rv->i;
			lv->unit = UNIT_NONE;
		} else if(rv->unit == UNIT_NONE) {
			lv->i /= rv->i;
		} else if(lv->unit == UNIT_NONE) {
			lv->i /= rv->i;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = (double)lv->i / ((double)rv->i * 25.4);
			else
				lv->d = (double)lv->i / ((double)rv->i / 25.4);
			lv->type = VAL_FLOAT;
			lv->unit = UNIT_NONE;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = (double)lv->i / ((double)rv->i * M_PI / 180.0);
			else
				lv->d = (double)lv->i / ((double)rv->i / M_PI * 180.0);
			lv->type = VAL_FLOAT;
			lv->unit = UNIT_NONE;
		} else
			rtinternal(ref, "Unit mismatch in op_div %d, %d", lv->unit, rv->unit);
		break;
	case VAL_FLOAT:
		assert(isscalar(rv));
		if(rv->type == VAL_INT)
			dval = rv->i;
		else
			dval = rv->d;
		if(fabs(dval) < EPSILON)
			rterror(ref, "Divide by zero (|val| < %e)", EPSILON);
		if(lv->unit == rv->unit || unitmismatch) {
			lv->d /= dval;
			lv->unit = UNIT_NONE;
		} else if(rv->unit == UNIT_NONE) {
			lv->d /= dval;
		} else if(lv->unit == UNIT_NONE) {
			lv->d /= dval;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d /= dval * 25.4;
			else
				lv->d /= dval / 25.4;
			lv->unit = UNIT_NONE;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d /= dval * M_PI / 180.0;
			else
				lv->d /= dval / M_PI * 180.0;
			lv->unit = UNIT_NONE;
		} else
			rtinternal(ref, "Unit mismatch in op_div %d, %d", lv->unit, rv->unit);
		break;
	case VAL_VECTOR:
		for(i = 0; i < lv->v.n; i++)
			op_div(ref, lv->v.vals[i], rv);
		break;
	case VAL_VECTORLIST:
		for(i = 0; i < lv->vl.n; i++)
			op_div(ref, lv->vl.vecs[i], rv);
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_div %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_mod(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	double dval;
	int unitmismatch = 0;
	assert(lv != NULL);
	assert(rv != NULL);

	/* Modulo "undef" = "undef" */
	if(lv->type == VAL_UNDEF)
		return lv;
	if(rv->type == VAL_UNDEF) {
		if(!isscalar(lv)) {
			rterror(ref, "Operation 'non-scalar %% <undef>' not supported");
			return lv;
		}
		lv->type = VAL_UNDEF;
		return lv;
	}

	/* Type check
	 * scalar % scalar
	 * vector % scalar
	 * vectorlist % scalar
	 */
	if(!isscalar(rv) || !(isscalar(lv) || isvector(lv) || isvectorlist(lv))) {
		rterror(ref, "Scalar modulo can only be calculated on scalars and vectors (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(isangle(lv) && isdistance(rv)) {
		rtwarning(ref, "Operation '%%': LHS has angular units and RHS has distance units, RHS units ignored");
		unitmismatch = 1;
	} else if(isdistance(lv) && isangle(rv)) {
		rtwarning(ref, "Operation '%%': LHS has distance units and RHS has angular units, RHS units ignored");
		unitmismatch = 1;
	}

	if(lv->type == VAL_INT && rv->type == VAL_FLOAT) {
		lv->d = lv->i;
		lv->type = VAL_FLOAT;
	}

	switch(lv->type) {
	case VAL_INT:
		assert(isscalar(rv));
		if(!rv->i)
			rterror(ref, "Modulo by zero");
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->i %= rv->i;
		} else if(lv->unit == UNIT_NONE) {
			lv->i %= rv->i;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = fmod((double)lv->i, ((double)rv->i * 25.4));
			else
				lv->d = fmod((double)lv->i, ((double)rv->i / 25.4));
			lv->type = VAL_FLOAT;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = fmod((double)lv->i, ((double)rv->i * M_PI / 180.0));
			else
				lv->d = fmod((double)lv->i, ((double)rv->i / M_PI * 180.0));
			lv->type = VAL_FLOAT;
		} else
			rtinternal(ref, "Unit mismatch in op_mod %d, %d", lv->unit, rv->unit);
		break;
	case VAL_FLOAT:
		assert(isscalar(rv));
		if(rv->type == VAL_INT)
			dval = rv->i;
		else
			dval = rv->d;
		if(fabs(dval) < EPSILON)
			rterror(ref, "Modulo by zero (|val| < %e)", EPSILON);
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->d = fmod(lv->d, dval);
		} else if(lv->unit == UNIT_NONE) {
			lv->d = fmod(lv->d, dval);
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = fmod(lv->d, (dval * 25.4));
			else
				lv->d = fmod(lv->d, (dval / 25.4));
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = fmod(lv->d, (dval * M_PI / 180.0));
			else
				lv->d = fmod(lv->d, (dval / M_PI * 180.0));
		} else
			rtinternal(ref, "Unit mismatch in op_mod %d, %d", lv->unit, rv->unit);
		break;
	case VAL_VECTOR:
		for(i = 0; i < lv->v.n; i++)
			op_mod(ref, lv->v.vals[i], rv);
		break;
	case VAL_VECTORLIST:
		for(i = 0; i < lv->vl.n; i++)
			op_mod(ref, lv->vl.vecs[i], rv);
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_mod %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_shl(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	int sft;
	assert(lv != NULL);
	assert(rv != NULL);

	if(lv->type == VAL_UNDEF || rv->type == VAL_UNDEF) {
		/* "undef" << x = undef */
		/* x << "undef" = x << 0 = x */
		return lv;
	}

	/* Type check
	 * scalar << scalar
	 * vector << scalar
	 * vectorlist << scalar
	 */
	if(!(isscalar(rv) && (isscalar(lv) || isvector(lv) || isvectorlist(lv)))) {
		rterror(ref, "Scalars, vectors and vectorlists can only be shifted by scalars (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(rv->unit != UNIT_NONE)
		rtwarning(ref, "Operation '<<': Shift count has units associated, ignored");
	if(rv->type == VAL_FLOAT && rv->d > (double)INT_MAX) {
		rterror(ref, "Operation '<<': Shift count floating point larger than integer can hold");
		return lv;
	}
	sft = value_to_int(rv);
	if(sft < 0) {
		rterror(ref, "Operation '<<': Shift count less than zero not allowed");
		return lv;
	}

	if(isscalar(lv) && sft > (int)(sizeof(int) * 8 - 1)) {
		rterror(ref, "Operation '<<': Shift count > %d has no effect", (int)(sizeof(int) * 8 - 1));
		return lv;
	} else if(!isscalar(lv) && sft > WARNSHIFTS)
		rtwarning(ref, "Operation '<<': Shift count > %d might be too large", sft);

	switch(lv->type) {
	case VAL_INT:
		lv->i <<= sft;
		break;
	case VAL_FLOAT:
		lv->d *= (double)(1 << sft);
		break;
	case VAL_VECTOR:
		for(i = 0; i < sft && lv->v.n; i++) {
			value_delete(lv->v.vals[0]);
			if(lv->v.n > 1)
				memmove(&lv->v.vals[0], &lv->v.vals[1], (lv->v.n-1)*sizeof(lv->v.vals[0]));
			lv->v.n--;
		}
		break;
	case VAL_VECTORLIST:
		for(i = 0; i < sft && lv->vl.n; i++) {
			value_delete(lv->vl.vecs[0]);
			if(lv->vl.n > 1)
				memmove(&lv->vl.vecs[0], &lv->vl.vecs[1], (lv->vl.n-1)*sizeof(lv->vl.vecs[0]));
			lv->vl.n--;
		}
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_shl %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_shr(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	int sft;
	assert(lv != NULL);
	assert(rv != NULL);

	if(lv->type == VAL_UNDEF || rv->type == VAL_UNDEF) {
		/* "undef" >> x = undef */
		/* x >> "undef" = x << 0 = x */
		return lv;
	}

	/* Type check
	 * scalar >> scalar
	 * vector >> scalar
	 * vectorlist > scalar
	 */
	if(!(isscalar(rv) && (isscalar(lv) || isvector(lv) || isvectorlist(lv)))) {
		rterror(ref, "Scalars, vectors and vectorlists can only be shifted by scalars (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(rv->unit != UNIT_NONE)
		rtwarning(ref, "Operation '>>': Shift count has units associated, ignored");
	if(rv->type == VAL_FLOAT && rv->d > (double)INT_MAX) {
		rterror(ref, "Operation '>>': Shift count floating point larger than integer can hold");
		return lv;
	}
	sft = value_to_int(rv);
	if(sft < 0) {
		rterror(ref, "Operation '>>': Shift count less than zero not allowed");
		return lv;
	}

	if(isscalar(lv) && sft > (int)(sizeof(int) * 8 - 1)) {
		rterror(ref, "Operation '>>': Shift count > %d has no effect", (int)(sizeof(int) * 8 - 1));
		return lv;
	} else if(!isscalar(lv) && sft > WARNSHIFTS)
		rtwarning(ref, "Operation '>>': Shift count > %d might be too large", sft);

	switch(lv->type) {
	case VAL_INT:
		lv->i >>= sft;
		break;
	case VAL_FLOAT:
		lv->d /= (double)(1 << sft);
		break;
	case VAL_VECTOR:
		for(i = 0; i < sft; i++)
			value_vector_prepend(lv, value_new(VAL_UNDEF));
		break;
	case VAL_VECTORLIST:
		for(i = 0; i < sft; i++)
			value_vectorlist_prepend(lv, value_new(VAL_VECTOR));
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_shl %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_call(node_t *n)
{
	value_t *v = NULL;
	int i, j;
	int nargs;

	assert(n->call.args == NULL || n->call.args->type == NT_EXPRLIST);

	nargs = n->call.args ? n->call.args->nlist.n : 0;

	/* Push all arguments onto the value stack */
	for(i = 0; i < nargs; i++)
		interpret_expr(n->call.args->nlist.nodes[i]);

	/* Check for builtin function */
	for(i = 0; i < nbuiltins; i++) {
		if(!strcmp(builtins[i].id, n->call.id)) {
			v = builtins[i].func(n, nargs);
			if(!v)
				rtinternal(n, "Builtin function returned no value (NULL)");
			for(i = 0; i < nargs; i++)
				value_delete(value_pop(n));
			return v;
		}
	}

	/* Script defined functions */
	for(i = 0; i < nfunctions; i++) {
		if(!strcmp(functions[i]->func.id, n->call.id)) {
			function_t *ft = &functions[i]->func;
			assert(!ft->args || ft->args->type == NT_ARGLIST);
			if((ft->args && ft->args->alist.n != nargs) || (!ft->args && nargs))
				rterror(n, "Incorrect number of arguments in call to '%s'", n->call.id);
			assert(ft->args == NULL || nargs == ft->args->alist.n);
			/* Open new scope and push the arguments as local variables */
			variable_scope_push();
			for(j = 0; j < nargs; j++)
				variable_set_local(ft->args->alist.ids[ft->args->alist.n - j - 1], value_pop(n));
			v = execute(ft->body, NULL);
			variable_scope_pop();
			return v;
		}
	}

	rterror(n, "Function '%s' not found", n->call.id);
	return NULL;
}

static value_t *op_indexref(const node_t *n, const variable_t **varref, int *iref)
{
	value_t *idx;
	value_t *val;
	const variable_t *var;
	int i;
	assert(varref != NULL);
	assert(iref != NULL);
	/* Indexed lvalue */
	idx = value_pop(n);	/* index */
	val = value_pop(n);	/* value to assign */
	if(!isscalar(idx)) {
		rterror(n, "Index on valalue must be scalar");
		value_delete(idx);
		value_delete(val);
		return NULL;
	}
	if(idx->unit != UNIT_NONE)
		rtwarning(n, "Index has units associated, ignored");
	i = value_to_int(idx);
	value_delete(idx);
	if(i < 0) {
		rterror(n, "Variable index on '%s' must be >= 0", n->expr.id);
		value_delete(val);
		return NULL;
	}
	if(!(var = variable_lookup(n->expr.id))) {
		rterror(n, "Variable '%s' undefined", n->expr.id);
		value_delete(val);
		return NULL;
	}
	if(isvector(var->v)) {
		if(!isscalar(val)) {
			rterror(n, "Rvalue must be scalar in vector indexed assignment");
			value_delete(val);
			return NULL;
		}
		if(i > 9)
			rtwarning(n, "Variable index on vector '%s' is larger than 9 (%d)", n->expr.id, i);
		/* Extend vector to fit index */
		while(var->v->v.n < i+1)
			value_vector_add_undef(var->v);
		*varref = var;
		*iref = i;
		return val;
	} else if(isvectorlist(var->v)) {
		if(!n->expr.dblidx && !isvector(val)) {
			rterror(n, "Rvalue must be vector in vectorlist indexed assignment");
			value_delete(val);
			return NULL;
		}
		/* Extend vectorlist to fit index */
		while(var->v->vl.n < i+1)
			value_vectorlist_add(var->v, value_new(VAL_VECTOR));
		*varref = var;
		*iref = i;
		return val;
	} else {
		rterror(n, "Variable '%s' must be vector or vectorlist for index", n->expr.id);
	}
	value_delete(val);
	return NULL;
}

static value_t *op_dblindex(const node_t *n, const variable_t *var, int idx, int *iref)
{
	int i;
	value_t *lv;
	value_t *vec;

	if(!n->expr.dblidx)
		return NULL;
	assert(var != NULL);
	assert(idx >= 0);
	assert(var->v->type == VAL_VECTORLIST);
	assert(var->v->vl.n > idx);

	interpret_expr(n->expr.dblidx);
	lv = value_pop(n);
	if(!isscalar(lv)) {
		rterror(n->expr.dblidx, "Secondary index must be scalar");
		return NULL;
	}
	if(lv->unit != UNIT_NONE)
		rtwarning(n->expr.dblidx, "Secondary index has units associated, ignored");
	i = value_to_int(lv);
	value_delete(lv);
	if(i > 9)
		rtwarning(n->expr.dblidx, "Secondary variable index on vector '%s' is larger than 9 (%d)", n->expr.id, i);

	vec = var->v->vl.vecs[idx];
	/* Extend vector to fit index */
	while(vec->v.n < i+1)
		value_vector_add_undef(vec);
	if(iref)
		*iref = i;
	return vec->v.vals[i];
}

static void op_xassign(const node_t *n, opfunc_t func)
{
	int i;
	const variable_t *var;
	value_t *lv;
	value_t *rv;
	if(n->expr.right) {
		if((lv = op_indexref(n, &var, &i))) {
			if(n->expr.dblidx && (!isvectorlist(var->v) || !isscalar(lv))) {
				if(!isscalar(lv))
					rterror(n, "Double index assignment must have scalar rvalue");
				else
					rterror(n, "Double index only allowed on vectorlist");
			} else if(isvector(var->v))
				func(n, var->v->v.vals[i], lv);
			else if(isvectorlist(var->v)) {
				if((rv = op_dblindex(n, var, i, NULL)))
					func(n, rv, lv);
				else
					func(n, var->v->vl.vecs[i], lv);
			}
			value_delete(lv);
		}
	} else {
		if(!(var = variable_lookup(n->expr.id))) {
			rterror(n, "Variable '%s' undefined", n->expr.id);
			value_delete(value_pop(n));
			return;
		}
		func(n, var->v, rv = value_pop(n));
		value_delete(rv);
	}
}

static void interpret_expr(node_t *n)
{
	int i, j;
	value_t *lv;
	value_t *rv;
	int lvb, rvb;
	double lvd, rvd;
	const variable_t *var;

	if(!n)
		return;

	if(n->type != NT_EXPR && n->type != NT_CALL)
		rtinternal(n, "Invalid node type %s (%d), expected NT_EXPR or NT_CALL", nodetypename(n->type), n->type);

	if(n->type == NT_CALL) {
		if(!(lv = op_call(n)))
			rterror(n, "Function did not return a value");
		value_push(lv);
		return;
	}

	interpret_expr(n->expr.left);
	/* Skip the right hand side initially on boolean ops to do short-circuit evaluation */
	if(n->expr.op != OP_LAND && n->expr.op != OP_LOR)
		interpret_expr(n->expr.right);
	switch(n->expr.op) {
	case OP_ADD:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_add(n, lv, rv));
		value_delete(rv);
		break;
	case OP_SUB:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_sub(n, lv, rv));
		value_delete(rv);
		break;
	case OP_MUL:
		rv = value_pop(n);
		lv = value_pop(n);
		/* scalar * vector[list] --> vector[list] * scalar */
		if(isscalar(lv) && (isvector(rv) || isvectorlist(rv))) {
			value_push(op_mul(n, rv, lv));
			value_delete(lv);
		} else {
			value_push(op_mul(n, lv, rv));
			value_delete(rv);
		}
		break;
	case OP_SHL:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_shl(n, lv, rv));
		value_delete(rv);
		break;
	case OP_SHR:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_shr(n, lv, rv));
		value_delete(rv);
		break;
	case OP_DIV:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_div(n, lv, rv));
		value_delete(rv);
		break;
	case OP_MOD:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_mod(n, lv, rv));
		value_delete(rv);
		break;
	case OP_INDEX:
		/* Left is value to index, right is index */
		rv = value_pop(n);
		lv = value_pop(n);
		if(!isvector(lv) && !isvectorlist(lv))
			rterror(n, "Source for [] operator is not vector or vectorlist");
		if(!isscalar(rv))
			rterror(n, "Index of [] operator is not scalar");
		if(rv->unit != UNIT_NONE)
			rtwarning(n, "Index has a unit (%s), ignored", unitname(rv->unit));

		i = value_to_int(rv);
		if(isvector(lv)) {
			if(i < 0 || i >= lv->v.n) {
				if(lv->v.n)
					rtwarning(n, "Vector index out of range (%d outside [0..%d])", i, lv->v.n-1);
				else
					rtwarning(n, "Vector index out of range (vector has no entries)");
				value_push(value_new(VAL_UNDEF));
			} else
				value_push(value_dup(lv->v.vals[i]));
		} else { 
			if(i < 0 || i >= lv->vl.n) {
				if(lv->vl.n)
					rtwarning(n, "Vectorlist index out of range (%d outside [0..%d])", i, lv->vl.n-1);
				else
					rtwarning(n, "Vectorlist index out of range (vectorlist has no entries)");
				value_push(value_new(VAL_VECTOR));
			} else
				value_push(value_dup(lv->vl.vecs[i]));
		}
		break;
	case OP_LOR:
	case OP_LAND:
		/* Only LHS got evaluated for logical ops */
		lv = value_pop(n);
		lvb = value_to_boolean(lv);
		value_delete(lv);
		/* Short-circuit the boolean operation */
		if(n->expr.op == OP_LAND && !lvb) {
			/* false && <anything> is always false */
			value_push(lv = value_new(VAL_INT));
			lv->i = 0;
			break;
		} else if(n->expr.op == OP_LOR && lvb) {
			/* true || <anything> is always true */
			value_push(lv = value_new(VAL_INT));
			lv->i = 1;
			break;
		}
		/* Undecided on LHS, evaluate RHS */
		interpret_expr(n->expr.right);
		rv = value_pop(n);
		rvb = value_to_boolean(rv);
		value_delete(rv);
		value_push(lv = value_new(VAL_INT));
		switch(n->expr.op) {
		case OP_LOR:	lv->i = lvb || rvb; break;
		case OP_LAND:	lv->i = lvb && rvb; break;
		default:	lv->i = 0; break;
		}
		break;
	case OP_EQ:
	case OP_NE:
	case OP_GT:
	case OP_LT:
	case OP_GE:
	case OP_LE:
		rv = value_pop(n);
		lv = value_pop(n);
		if(lv->unit != rv->unit)
			rtwarning(n, "Units in comparison not equal, converted to mm (LHS=%s, RHS=%s)", unitname(lv->unit), unitname(rv->unit));
		rvd = value_to_mm(rv);
		lvd = value_to_mm(lv);
		value_delete(rv);
		value_delete(lv);
		value_push(lv = value_new(VAL_INT));
		switch(n->expr.op) {
		case OP_EQ:	lv->i = fabs(lvd - rvd) < EPSILON; break;
		case OP_NE:	lv->i = fabs(lvd - rvd) >= EPSILON; break;
		case OP_GT:	lv->i = lvd > rvd; break;
		case OP_LT:	lv->i = lvd < rvd; break;
		case OP_GE:	lv->i = lvd >= rvd; break;
		case OP_LE:	lv->i = lvd <= rvd; break;
		default:	lv->i = 0; break;
		}
		break;
	case OP_NOT:
		lv = value_pop(n);
		lvb = value_to_boolean(lv);
		value_delete(lv);
		value_push(lv = value_new(VAL_INT));
		lv->i = !lvb;
		break;
	case OP_INT:
		value_push(value_new_int(n->expr.i, n->expr.unit));
		break;
	case OP_FLOAT:
		value_push(value_new_flt(n->expr.d, n->expr.unit));
		break;
	case OP_STRING:
		value_push(value_new_str(n->expr.id));
		break;
	case OP_VECTOR:
		lv = value_new(VAL_VECTOR);
		for(i = 0; i < n->expr.nlist.n; i++) {
			if(!n->expr.nlist.nodes[i])
				value_vector_add_undef(lv);
			else {
				interpret_expr(n->expr.nlist.nodes[i]);
				value_vector_add(lv, value_pop(n->expr.nlist.nodes[i]));
			}
		}
		value_push(lv);
		break;
	case OP_VECTORLIST:
		lv = value_new(VAL_VECTORLIST);
		for(i = 0; i < n->expr.nlist.n; i++) {
			interpret_expr(n->expr.nlist.nodes[i]);
			rv = value_pop(n->expr.nlist.nodes[i]);
			assert(rv != NULL);
			if(rv->type != VAL_VECTOR)
				rterror(n, "Vectorlist entries must be vectors");
			value_vectorlist_add(lv, rv);
		}
		value_push(lv);
		break;
	case OP_ASSIGN:
		if(n->expr.right) {
			if((lv = op_indexref(n, &var, &i))) {
				if(n->expr.dblidx && (!isvectorlist(var->v) || !isscalar(lv))) {
					if(!isscalar(lv))
						rterror(n, "Double index assignment must have scalar rvalue");
					else
						rterror(n, "Double index only allowed on vectorlist");
				} else if(isvector(var->v)) {
					value_delete(var->v->v.vals[i]);
					var->v->v.vals[i] = lv;
				} else if(isvectorlist(var->v)) {
					if((rv = op_dblindex(n, var, i, &j))) {
						value_delete(var->v->vl.vecs[i]->v.vals[j]);
						var->v->vl.vecs[i]->v.vals[j] = lv;
					} else {
						value_delete(var->v->vl.vecs[i]);
						var->v->vl.vecs[i] = lv;
					}
				}
			}
		} else {
			variable_set(n->expr.id, value_pop(n));
		}
		break;
	case OP_ADDASSIGN:
		op_xassign(n, op_add);
		break;
	case OP_SUBASSIGN:
		op_xassign(n, op_sub);
		break;
	case OP_MULASSIGN:
		op_xassign(n, op_mul);
		break;
	case OP_DIVASSIGN:
		op_xassign(n, op_div);
		break;
	case OP_MODASSIGN:
		op_xassign(n, op_mod);
		break;
	case OP_SHLASSIGN:
		op_xassign(n, op_shl);
		break;
	case OP_SHRASSIGN:
		op_xassign(n, op_shr);
		break;
	case OP_INC:
		if(!(var = variable_lookup(n->expr.id))) {
			rterror(n, "Variable '%s' undefined", n->expr.id);
			break;
		}
		switch(var->v->type) {
		case VAL_INT:	var->v->i++; break;
		case VAL_FLOAT:	var->v->d += 1.0; break;
		default:
			rtwarning(n, "Increment on non-scalar has no effect");
			break;
		}
		break;
	case OP_DEC:
		if(!(var = variable_lookup(n->expr.id))) {
			rterror(n, "Variable '%s' undefined", n->expr.id);
			break;
		}
		switch(var->v->type) {
		case VAL_INT:	var->v->i--; break;
		case VAL_FLOAT:	var->v->d -= 1.0; break;
		default:
			rtwarning(n, "Increment on non-scalar has no effect");
			break;
		}
		break;
	case OP_DEREF:
		if(!(var = variable_lookup(n->expr.id))) {
			rterror(n, "Variable '%s' undefined", n->expr.id);
			value_push(value_new(VAL_UNDEF));
			break;
		}
		value_push(value_dup(var->v));
		break;
	default:
		rtinternal(n, "Invalid expression operation %s (%d)", expropname(n->expr.op), n->expr.op);
		break;
	}
}

static value_t *op_foreach(node_t *n)
{
	int i;
	value_t *v = NULL;
	value_t *src;
	node_t *brk = NULL;
	assert(n->lfe.src->type == NT_EXPR);
	interpret_expr(n->lfe.src);
	src = value_pop(n);
	if(!isvectorlist(src))
		rterror(n, "Foreach's source is not a vectorlist");
	for(i = 0; i < src->vl.n; i++) {
		value_t *val = value_dup(src->vl.vecs[i]);
		assert(val->type == VAL_VECTOR);
		variable_set(n->lfe.dst, val);
		v = execute(n->lfe.stmts, &brk);
		if(v)
			break;
		if(brk && brk->type == NT_BREAK)
			break;
		if(brk && brk->type == NT_CONTINUE)
			brk = NULL;
	}
	value_delete(src);
	return v;
}

static void function_add(node_t *n)
{
	int i;
	assert(n->type == NT_FUNCTION);
	testalloc((void **)&functions, nfunctions, &nafunctions, sizeof(functions[0]));
	for(i = 0; i < nfunctions; i++) {
		if(!strcmp(functions[i]->func.id, n->func.id))
			rtinternal(n, "Function '%s' already defined", n->func.id);
	}
	for(i = 0; i < nbuiltins; i++) {
		if(!strcmp(builtins[i].id, n->func.id))
			rtinternal(n, "Function '%s' is a builtin function", n->func.id);
	}
	functions[nfunctions] = n;
	nfunctions++;
}

static value_t *op_for(node_t *n)
{
	node_t *brk = NULL;
	value_t *v;
	int b;
	assert(n->lfor.cond->type == NT_EXPR);

	if(execute(n->lfor.init, NULL))
		rterror(n, "For-loop initialization should not return a value");
	while(1) {
		interpret_expr(n->lfor.cond);
		v = value_pop(n);
		b = value_to_boolean(v);
		value_delete(v);
		if(!b)
			break;
		if((v = execute(n->lfor.stmts, &brk)))
			return v;
		if(brk && brk->type == NT_BREAK)
			break;
		if(brk && brk->type == NT_CONTINUE)
			brk = NULL;
		if(execute(n->lfor.inc, NULL))
			rterror(n, "For-loop increment should not return a value");
	}
	return NULL;
}

static value_t *op_while(node_t *n)
{
	node_t *brk = NULL;
	value_t *v;
	int b;
	assert(n->lfor.cond->type == NT_EXPR);

	while(1) {
		interpret_expr(n->lfor.cond);
		v = value_pop(n);
		b = value_to_boolean(v);
		value_delete(v);
		if(!b)
			break;
		if((v = execute(n->lfor.stmts, &brk)))
			return v;
		if(brk && brk->type == NT_BREAK)
			break;
		if(brk && brk->type == NT_CONTINUE)
			brk = NULL;
	}
	return NULL;
}

static value_t *op_dowhile(node_t *n)
{
	node_t *brk = NULL;
	value_t *v;
	int b;
	assert(n->lfor.cond->type == NT_EXPR);

	while(1) {
		if((v = execute(n->lfor.stmts, &brk)))
			return v;
		if(brk && brk->type == NT_BREAK)
			break;
		if(brk && brk->type == NT_CONTINUE)
			brk = NULL;
		else {
			interpret_expr(n->lfor.cond);
			v = value_pop(n);
			b = value_to_boolean(v);
			value_delete(v);
			if(!b)
				break;
		}
	}
	return NULL;
}

static value_t *op_if(node_t *n, node_t **brkref)
{
	value_t *v;
	int b;
	assert(n->cond.cond->type == NT_EXPR);

	interpret_expr(n->cond.cond);
	v = value_pop(n);
	b = value_to_boolean(v);
	value_delete(v);
	if(b) {
		if((v = execute(n->cond.ifclause, brkref)))
			return v;
	} else if(n->cond.elseclause) {
		if((v = execute(n->cond.elseclause, brkref)))
			return v;
	}
	return NULL;
}

static value_t *interpret(node_t *n, node_t **brkref)
{
	value_t *v;
	int i;
	if(!n)
		return NULL;

	switch(n->type) {
	case NT_EXPRLIST:	/* Only as call argument */
	case NT_ARGLIST:	/* Only as function arg definition */
	case NT_BREAK:		/* Only in loops */
	case NT_CONTINUE:	/* Only in loops */
	default:
		rtinternal(n, "Invalid node type %s (%d)", nodetypename(n->type), n->type);
		break;
	case NT_EXPR:		interpret_expr(n); break;
	case NT_IF:		return op_if(n, brkref);
	case NT_FOR:		return op_for(n);
	case NT_FOREACH:	return op_foreach(n);
	case NT_WHILE:		return op_while(n);
	case NT_DOWHILE:	return op_dowhile(n);
	case NT_FUNCTION:	function_add(n); break;
	case NT_CALL:
		if((v = op_call(n)))
			value_delete(v);
		break;
	case NT_RETURN:
		if(n->eref) {
			interpret_expr(n->eref);
			return value_pop(n);
		} else
			return value_new(VAL_UNDEF);
	case NT_LOCAL:
		assert(n->eref != NULL && n->eref->type == NT_ARGLIST);
		for(i = 0; i < n->eref->alist.n; i++)
			variable_set_local(n->eref->alist.ids[i], value_new(VAL_UNDEF));
		break;
	}
	return NULL;
}

static value_t *execute(node_t *head, node_t **brkref)
{
	node_t *n;
	value_t *v;
	for(n = head; n; n = n->next) {
		if(n->type == NT_BREAK || n->type == NT_CONTINUE) {
			if(brkref) {
				*brkref = n;
				return NULL;
			} else
				rterror(n, "Unexpected break or continue");
		}
		if((v = interpret(n, brkref)))
			return v;
		if(brkref && *brkref)
			return NULL;
	}
	return NULL;
}

#define BUFSIZE	16384
static void cpyfile(const char *src, FILE *fp)
{
	FILE *sp = fopen(src, "rb");
	char *buf;
	if(!sp) {
		rterror(NULL, "File '%s' cannot be opened", src);
		return;
	}
	buf = malloc(BUFSIZE);
	assert(buf != NULL);
	while(1) {
		size_t n = fread(buf, 1, BUFSIZE, sp);
		if(!n)
			break;
		while(n) {
			size_t w = fwrite(buf, 1, n, fp);
			if(!n) {
				if(ferror(fp)) {
					rterror(NULL, "Error writing output file");
					goto err_out;
				}
			} else
				n -= w;
		}
	}
err_out:
	fclose(sp);
	free(buf);
}

static void prologue(const char *src, FILE *fp)
{
	if(src)
		cpyfile(src, fp);
	else {
		char str[64];
		time_t t;
		struct tm *tm;
		t = time(NULL);
		tm = localtime(&t);
		strftime(str, sizeof(str), "%Y-%m-%d %H:%M:%S", tm);
		fprintf(fp, "(gcmc compiled code, do not change)\n");
		fprintf(fp, "(%s)\n", str);
		fprintf(fp, "(-- prologue begin --)\n");
		switch(global_plane) {
		case PLANE_XY:	fprintf(fp, "G17 ( Use XY plane )\n"); break;
		case PLANE_XZ:	fprintf(fp, "G18 ( Use XZ plane )\n"); break;
		case PLANE_YZ:	fprintf(fp, "G19 ( Use YZ plane )\n"); break;
		}
		if(cl_inch)
			fprintf(fp, "G20 ( Use inch )\n");
		else
			fprintf(fp, "G21 ( Use mm )\n");
		fprintf(fp, "G40 ( Cancel cutter radius compensation )\n");
		fprintf(fp, "G49 ( Cancel tool length compensation )\n");
		fprintf(fp, "G54 ( Default coordinate system )\n");
		fprintf(fp, "G80 ( Cancel canned cycle )\n");
		if(cl_relative)
			fprintf(fp, "G91 ( Use relative distance mode )\n");
		else
			fprintf(fp, "G90 ( Use absolute distance mode )\n");
		fprintf(fp, "G94 ( Units Per Minute feed rate mode )\n");
		fprintf(fp, "(-- prologue end --)\n");
	}
}

static void epilogue(const char *src, FILE *fp)
{
	if(src)
		cpyfile(src, fp);
	else {
		fprintf(fp, "(-- epilogue begin --)\n");
		fprintf(fp, "M30 (-- epilogue end --)\n");
	}
}

static const char **includepaths;
static int nincludepaths;
static int naincludepaths;

static void includepath_add(const char *path)
{
	assert(path != NULL);
	testalloc((void **)&includepaths, nincludepaths, &naincludepaths, sizeof(*includepaths));
	includepaths[nincludepaths] = path;
	nincludepaths++;
}

#ifdef FREEONEXIT
static void includepath_purge(void)
{
	int i;
	for(i = 0; i < nincludepaths; i++)
		free((void *)includepaths[i]);
	free(includepaths);
}
#endif

const char *includepath_get(int idx)
{
	if(idx < 0)
		rtinternal(NULL, "Negative index in index search path retrieval");
	if(idx >= nincludepaths)
		return NULL;
	return includepaths[idx];
}

static const char usage_str[] =
	"gcmc [options] [inputfile]\n"
	"Options:\n"
	"  -a|--offset-a ofs   Set global A-offset\n"
	"  -b|--offset-b ofs   Set global B-offset\n"
	"  -c|--offset-c ofs   Set global C-offset\n"
	"  -d|--debug          Enable debug\n"
	"  -G|--prologue file  Use 'file' as prologue code\n"
	"  -g|--epilogue file  Use 'file' as epilogue code\n"
	"  -h|--help           This message\n"
	"  -I|--include        path  Add include search path\n"
	"  -i|--imperial       Use imperial output units (default mm)\n"
	"  -o|--output file    Output to 'file' (default stdout)\n"
	"  -P|--precision dec  Use 'dec' decimals in output\n"
	"  -p|--plane plane    Set initial cutting plane (0|XY, 1|XZ, 2|YZ)\n"
	"  -q|--bare           Do not output prologue/epilogue\n"
	"  -r|--relative       Use relative positioning\n"
	"  -U|--uvw            Enable 9-axis base XYZABCUVW (default XYZABC)\n"
	"  -u|--offset-u ofs   Set global U-offset\n"
	"  -V|--version        Print version and exit\n"
	"  -v|--offset-v ofs   Set global V-offset\n"
	"  -W|--error          Exit non-zero if warnings are encountered\n"
	"  -w|--offset-w ofs   Set global W-offset\n"
	"  -x|--offset-x ofs   Set global X-offset\n"
	"  -y|--offset-y ofs   Set global Y-offset\n"
	"  -z|--offset-z ofs   Set global Z-offset\n"
	;

static const struct option lopts[] = {
	{ "debug",	no_argument,		NULL,	'd' },
	{ "help",	no_argument,		NULL,	'h' },
	{ "include",	required_argument,	NULL,	'I' },
	{ "imperial",	no_argument,		NULL,	'i' },
	{ "output",	required_argument,	NULL,	'o' },
	{ "precision",	required_argument,	NULL,	'P' },
	{ "plane",	required_argument,	NULL,	'p' },
	{ "bare",	no_argument,		NULL,	'q' },
	{ "relative",	no_argument,		NULL,	'r' },
	{ "uvw",	no_argument,		NULL,	'U' },
	{ "version",	no_argument,		NULL,	'V' },
	{ "error",	no_argument,		NULL,	'W' },
	{ "prologue",	required_argument,	NULL,	'G' },
	{ "epilogue",	required_argument,	NULL,	'g' },
	{ "offset-x",	required_argument,	NULL,	'x' },
	{ "offset-y",	required_argument,	NULL,	'y' },
	{ "offset-z",	required_argument,	NULL,	'z' },
	{ "offset-a",	required_argument,	NULL,	'a' },
	{ "offset-b",	required_argument,	NULL,	'b' },
	{ "offset-c",	required_argument,	NULL,	'c' },
	{ "offset-u",	required_argument,	NULL,	'u' },
	{ "offset-v",	required_argument,	NULL,	'x' },
	{ "offset-w",	required_argument,	NULL,	'x' },
	{ NULL,		0,			NULL,	 0 },
};

#define MAXOFFSET	1e6	/* 1000000mm is one kilometer, should be enough for a mill... */
#define NOFFSETS	9	/* XYZABCUVW */

static int get_offset(double *offs, const char *str, char ch)
{
	char *cptr;
	assert(offs != NULL);
	assert(str != NULL);
	*offs = strtod(str, &cptr);
	if(*cptr) {
		fprintf(stderr, "Invalid global %c-offset\n", ch);
		return 0;
	} else if(fabs(*offs) > MAXOFFSET) {
		fprintf(stderr, "Global %c-offset out of range\n", ch);
		return 0;
	}
	return 1;
}

int main(int argc, char *argv[])
{
	int i;
	int unit;
	int anyoffs;
	int debug = 0;
	int lose = 0;
	int optc;
	int oidx = 0;
	char *cptr;
	const char *ofn = NULL;
	char *profn = NULL;
	char *epifn = NULL;
	int retval = 0;
	value_t *v, *w;
	double offsets[NOFFSETS] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
	int proepi = 1;
	int warnerror = 0;

	while(EOF != (optc = getopt_long(argc, argv, "a:b:c:dG:g:hI:io:p:P:qrUVWu:v:w:x:y:z:", lopts, &oidx))) {
		switch(optc) {
		case 'a':
			if(!get_offset(&offsets[3], optarg, 'A'))
				lose++;
			break;
		case 'b':
			if(!get_offset(&offsets[4], optarg, 'B'))
				lose++;
			break;
		case 'c':
			if(!get_offset(&offsets[5], optarg, 'C'))
				lose++;
			break;
		case 'd':
			debug = 1;
			break;
		case 'h':
			fprintf(stderr, "%s", usage_str);
			exit(1);
			break;
		case 'G':
			profn = strdup(optarg);
			break;
		case 'g':
			epifn = strdup(optarg);
			break;
		case 'I':
			includepath_add(strdup(optarg));
			break;
		case 'i':
			cl_inch = 1;
			break;
		case 'o':
			ofn = optarg;
			break;
		case 'p':
			global_plane = strtoul(optarg, &cptr, 10);
			if(*cptr) {
				if(!strcasecmp(optarg, "xy"))
					global_plane = PLANE_XY;
				else if(!strcasecmp(optarg, "xz"))
					global_plane = PLANE_XZ;
				else if(!strcasecmp(optarg, "yz"))
					global_plane = PLANE_YZ;
				else
					goto plane_error;
			} else if(global_plane < 0 || global_plane > 2) {
plane_error:
				fprintf(stderr, "Initial plane must be 0, 1, 2, XY, XZ or YZ\n");
				lose++;
			}
			break;
		case 'P':
			cl_decimals = strtoul(optarg, &cptr, 10);
			if(*cptr) {
				fprintf(stderr, "Invalid number of decimals\n");
				lose++;
			}
			if(cl_decimals < 1 || cl_decimals > 15) {
				fprintf(stderr, "Number of decimals out of range (1..15)\n");
				lose++;
			}
			if(cl_decimals < 3)
				fprintf(stderr, "Number of decimals less than 3 severely limits accuracy\n");
			break;
		case 'q':
			proepi = 0;
			break;
		case 'r':
			cl_relative = 1;
			break;
		case 'x':
			if(!get_offset(&offsets[0], optarg, 'X'))
				lose++;
			break;
		case 'y':
			if(!get_offset(&offsets[1], optarg, 'Y'))
				lose++;
			break;
		case 'z':
			if(!get_offset(&offsets[2], optarg, 'Z'))
				lose++;
			break;
		case 'u':
			if(!get_offset(&offsets[3], optarg, 'U'))
				lose++;
			break;
		case 'v':
			if(!get_offset(&offsets[4], optarg, 'V'))
				lose++;
			break;
		case 'w':
			if(!get_offset(&offsets[5], optarg, 'W'))
				lose++;
			break;
		case 'U':
			naxes = 9;
			break;
		case 'V':
			printf("G-Code Meta Compiler - gcmc version %s\n", PACKAGE_VERSION);
			printf("Distributed under GPLv3+\n");
			lose++;
			break;
		case 'W':
			warnerror = 1;
			break;
		case '?':
			lose++;
			break;
		default:
			fprintf(stderr, "Unknown option character '%c' (%d) in option handling\n", isprint(optc) ? optc : ' ', optc);
			lose++;
			break;
		}
	}

	for(i = anyoffs = 0; i < NOFFSETS; i++)
		anyoffs |= offsets[i] != 0.0;
	if(cl_relative && anyoffs) {
		fprintf(stderr, "Offsets must be zero when using relative addressing\n");
		lose++;
	}

	if(lose)
		exit(1);

	if(debug) {
		setbuf(stdout, NULL);
		setbuf(stderr, NULL);
		yydebug = 1;
	}

	/* Last search-path entry is current directory */
	includepath_add(strdup("."));

	if(optind >= argc) {
		yyin = stdin;
		filename = strdup("--stdin--");
	} else {
		if(NULL == (yyin = fopen(argv[optind], "r"))) {
			perror(argv[optind]);
			exit(1);
		}
		filename = strdup(argv[optind]);
	}

	if(!ofn) {
		ofp = stdout;
	} else {
		if(NULL == (ofp = fopen(ofn, "w"))) {
			perror(ofn);
			exit(1);
		}
	}

	assert(naxes == 6 || naxes == 9);

	unit = cl_inch ? UNIT_IN : UNIT_MM;

	/* Setup reference to global offset and position track valiable */
	v = value_new(VAL_VECTOR);
	for(i = 0; i < naxes; i++) {
		w = value_new_flt(offsets[i], unit);
		value_vector_add(v, w);
		global_offs[i] = &w->d;
	}
	variable_set("global_offset", v);

	v = value_new(VAL_VECTOR);
	w = value_new_flt(0.0, unit); value_vector_add(v, w); global_pos[0] = &w->d;
	w = value_new_flt(0.0, unit); value_vector_add(v, w); global_pos[1] = &w->d;
	w = value_new_flt(0.0, unit); value_vector_add(v, w); global_pos[2] = &w->d;
	w = value_new_flt(0.0, UNIT_RAD); value_vector_add(v, w); global_pos[3] = &w->d;
	w = value_new_flt(0.0, UNIT_RAD); value_vector_add(v, w); global_pos[4] = &w->d;
	w = value_new_flt(0.0, UNIT_RAD); value_vector_add(v, w); global_pos[5] = &w->d;
	if(naxes > 6) {
		w = value_new_flt(0.0, unit); value_vector_add(v, w); global_pos[6] = &w->d;
		w = value_new_flt(0.0, unit); value_vector_add(v, w); global_pos[7] = &w->d;
		w = value_new_flt(0.0, unit); value_vector_add(v, w); global_pos[8] = &w->d;
	}
	variable_set("global_position", v);

	/* Parse the input script */
	yyparse();

	if(syntaxerrors)
		exit(1);

	/* Output the G-code */
	if(proepi || profn)
		prologue(profn, ofp);
	if((v = execute(scripthead, NULL))) {
		retval = value_to_int(v);
		value_delete(v);
	}
	if(proepi || epifn)
		epilogue(epifn, ofp);

	if(warnerror && runtimewarnings)
		rterror(NULL, "Runtime warnings considered to be errors");

	if(runtimeerrors)
		retval = 1;

	/* Cleanup for testing purposes */
	assert(0 == value_stackdepth());

#ifdef FREEONEXIT
	variable_purge();
	value_purge();
	node_delete(scripthead);
	if(functions)
		free(functions);
	includepath_purge();
	free((void *)filename);
#endif
	fclose(ofp);

	return retval;
}
